@@ -18,16 +18,16 @@
 
 import bz2
 import collections
+import json
 import logging
 import os
 import pickle
 import sys
 
 import numpy
-import skimage
 
-from imagecat.color import linear_to_srgb, srgb_to_linear
-from imagecat.data import Image, Layer, Role
+import imagecat.color
+import imagecat.data
 
 try:
     import Imath
@@ -68,8 +68,26 @@ def openexr_loader(task, path, layers):
     width = header["dataWindow"].max.x - header["dataWindow"].min.x + 1
     height = header["dataWindow"].max.y - header["dataWindow"].min.y + 1
 
+    # Cleanup the OpenEXR header to use as the image metadata.
+    class CustomEncoder(json.JSONEncoder):
+        def default(self, o):
+            if isinstance(o, bytes):
+                return o.decode("UTF-8")
+            elif isinstance(o, Imath.Channel):
+                return (o.type, o.xSampling, o.ySampling)
+            elif isinstance(o, (Imath.Compression, Imath.LineOrder, Imath.PixelType)):
+                return str(o)
+            elif isinstance(o, (Imath.V2f, Imath.V2i, Imath.point)):
+                return (o.x, o.y)
+            elif isinstance(o, (Imath.Box, Imath.Box2f, Imath.Box2i)):
+                return (o.min.x, o.min.y, o.max.x, o.max.y)
+            return super().default(o)
+
+    metadata = json.dumps(header, cls=CustomEncoder)
+    metadata = json.loads(metadata)
+
     # Load each OpenEXR channel into a separate Imagecat Layer.
-    image = Image()
+    layers = {}
     for name, dtype in header["channels"].items():
         if dtype.type.v == Imath.PixelType.HALF:
             data = numpy.frombuffer(reader.channel(name), dtype=numpy.float16).reshape((height, width, 1))
@@ -77,8 +95,9 @@ def openexr_loader(task, path, layers):
             data = numpy.frombuffer(reader.channel(name), dtype=numpy.float32).reshape((height, width, 1))
         elif dtype.type.v == Imath.PixelType.INT:
             data = numpy.frombuffer(reader.channel(name), dtype=numpy.int32).reshape((height, width, 1))
-        image.layers[name] = Layer(data=data, components=[""], role=Role.NONE)
-    return image
+        layers[name] = imagecat.data.Layer(data=data, role=imagecat.data.Role.NONE)
+
+    return imagecat.data.Image(layers=layers, metadata=metadata)
 
 
 def pickle_loader(task, path, layers):
@@ -99,7 +118,7 @@ def pickle_loader(task, path, layers):
 
     with bz2.open(path, "rb") as stream:
         image = pickle.load(stream)
-    if not isinstance(image, Image):
+    if not isinstance(image, imagecat.data.Image):
         raise RuntimeError("Not an Imagecat Pickle (*.icp) file.") # pragma: no cover
     return image
 
@@ -120,14 +139,14 @@ def pil_loader(task, path, layers):
     pil_image = PIL.Image.open(path)
     log.debug(pil_image.info)
 
-    image = Image()
+    image = imagecat.data.Image()
     if pil_image.mode == "L":
-        image.layers["Y"] = Layer(data=numpy.array(pil_image, dtype=numpy.float16) / 255.0)
+        image.layers["Y"] = imagecat.data.Layer(data=numpy.array(pil_image, dtype=numpy.float16) / 255.0)
     if pil_image.mode == "RGB":
-        image.layers["C"] = Layer(data=srgb_to_linear(numpy.array(pil_image, dtype=numpy.float16) / 255.0), role=Role.RGB)
+        image.layers["C"] = imagecat.data.Layer(data=imagecat.color.srgb_to_linear(numpy.array(pil_image, dtype=numpy.float16) / 255.0), role=imagecat.data.Role.RGB)
     if pil_image.mode == "RGBA":
-        image.layers["C"] = Layer(data=srgb_to_linear(numpy.array(pil_image, dtype=numpy.float16)[:,:,0:3] / 255.0), role=Role.RGB)
-        image.layers["A"] = Layer(data=numpy.array(pil_image, dtype=numpy.float16)[:,:,3:4] / 255.0)
+        image.layers["C"] = imagecat.data.Layer(data=imagecat.color.srgb_to_linear(numpy.array(pil_image, dtype=numpy.float16)[:,:,0:3] / 255.0), role=imagecat.data.Role.RGB)
+        image.layers["A"] = imagecat.data.Layer(data=numpy.array(pil_image, dtype=numpy.float16)[:,:,3:4] / 255.0)
     return image
 
 
@@ -148,28 +167,28 @@ def openexr_saver(task, image, layers, path):
     if extension != ".exr":
         return False
 
-    channels = {}
-    pixels = {}
+    exr_channels = {}
+    exr_pixels = {}
     for layer_name in layers:
         layer = image.layers[layer_name]
         dtype = layer.data.dtype
         shape = layer.data.shape
-        for index, component in enumerate(layer.components):
-            channel_name = f"{layer_name}.{component}" if component else f"{layer_name}"
+        for index, channel in enumerate(imagecat.data.channels(layer.role)):
+            channel_name = f"{layer_name}.{channel}" if channel else f"{layer_name}"
             if dtype == numpy.float16:
-                channels[channel_name] = Imath.Channel(Imath.PixelType(Imath.PixelType.HALF))
+                exr_channels[channel_name] = Imath.Channel(Imath.PixelType(Imath.PixelType.HALF))
             elif dtype == numpy.float32:
-                channels[channel_name] = Imath.Channel(Imath.PixelType(Imath.PixelType.FLOAT))
+                exr_channels[channel_name] = Imath.Channel(Imath.PixelType(Imath.PixelType.FLOAT))
             elif dtype == numpy.int32:
-                channels[channel_name] = Imath.Channel(Imath.PixelType(Imath.PixelType.INT))
+                exr_channels[channel_name] = Imath.Channel(Imath.PixelType(Imath.PixelType.INT))
             else:
                 raise ValueError(f"Unsupported dtype: {dtype}")
-            pixels[channel_name] = layer.data[:,:,index].tobytes()
+            exr_pixels[channel_name] = layer.data[:,:,index].tobytes()
 
-    header = OpenEXR.Header(shape[1], shape[0])
-    header["channels"] = channels
-    writer = OpenEXR.OutputFile(path, header)
-    writer.writePixels(pixels)
+    exr_header = OpenEXR.Header(shape[1], shape[0])
+    exr_header["channels"] = exr_channels
+    exr_writer = OpenEXR.OutputFile(path, exr_header)
+    exr_writer.writePixels(exr_pixels)
 
     return True
 
@@ -202,21 +221,10 @@ def pil_saver(task, image, layers, path):
     if "PIL.Image" not in sys.modules:
         return False # pragma: no cover
 
-    if len(layers) == 1:
-        layer = image.layers[layers[0]]
-        if layer.data.shape[2] == 1:
-            pil_image = PIL.Image.fromarray(skimage.img_as_ubyte(layer.data[:,:,0]), mode="L")
-        elif layer.data.shape[2] == 3 and layer.role == Role.RGB:
-            pil_image = PIL.Image.fromarray(skimage.img_as_ubyte(linear_to_srgb(layer.data)), mode="RGB")
-        else:
-            return False
-    elif len(layers) == 2 and "C" in layers and "A" in layers:
-        C = linear_to_srgb(image.layers["C"].data)
-        A = image.layers["A"].data
-        pil_image = PIL.Image.fromarray(skimage.img_as_ubyte(numpy.dstack((C, A))), mode="RGBA")
-    else:
+    if len(layers) != 1:
         return False
 
+    pil_image = imagecat.data.to_pil(image.layers[layers[0]])
     pil_image.save(path)
     return True
 
